@@ -7,15 +7,13 @@ from sqliteparser.ast import OnConflict, OnDelete, OnUpdate
 
 from . import columns, migrations
 from .database import (
-    AutoTable,
     ColumnDoesNotExistError,
     Database,
     ISqliteApiError,
     ISqliteError,
-    PrintDebugger,
-    Table,
     TableDoesNotExistError,
 )
+from .schema import AutoTable, Schema, Table
 
 
 def sqlite3_convert_boolean(b):
